@@ -234,10 +234,24 @@ bacon_video_widget_properties_update (BaconVideoWidgetProperties *props,
 		UPDATE_FROM_INT2 (BVW_INFO_DIMENSION_X, BVW_INFO_DIMENSION_Y,
 				  "dimensions", N_("%d x %d"));
 		UPDATE_FROM_STRING (BVW_INFO_VIDEO_CODEC, "vcodec");
-		UPDATE_FROM_INT (BVW_INFO_FPS, "framerate",
-				 N_("%d frames per second"), _("N/A"));
 		UPDATE_FROM_INT (BVW_INFO_VIDEO_BITRATE, "video_bitrate",
 				 N_("%d kbps"), _("N/A"));
+
+		/* The FPS has to be done differently because it's a plural string */
+		{
+			gchar *temp;
+			bacon_video_widget_get_metadata (BACON_VIDEO_WIDGET (bvw), BVW_INFO_FPS, &value);
+			if (g_value_get_int (&value) != 0) {
+				temp = g_strdup_printf (ngettext ("%d frame per second", "%d frames per second", g_value_get_int (&value)),
+				                        g_value_get_int (&value));
+			} else {
+				temp = g_strdup (_("N/A"));
+			}
+			bacon_video_widget_properties_set_label (props, "framerate", temp);
+			g_free (temp);
+			g_value_unset (&value);
+		}
+
 		gtk_widget_show (item);
 	} else {
 		gtk_widget_hide (item);
