@@ -68,8 +68,6 @@ typedef struct {
 
 struct TotemPlaylistPrivate
 {
-	GtkBuilder *xml;
-
 	GtkWidget *treeview;
 	GtkTreeModel *model;
 	GtkTreePath *current;
@@ -79,6 +77,12 @@ struct TotemPlaylistPrivate
 	GtkActionGroup *action_group;
 	GtkUIManager *ui_manager;
 
+	/* Widgets */
+	GtkWidget *save_button;
+	GtkWidget *remove_button;
+	GtkWidget *up_button;
+	GtkWidget *down_button;
+
 	/* These is the current paths for the file selectors */
 	char *path;
 	char *save_path;
@@ -95,6 +99,9 @@ struct TotemPlaylistPrivate
 	GtkTreePath *tree_path;
 	GtkTreeViewDropPosition drop_pos;
 
+	/* Cursor ref: 0 if the cursor is unbusy; positive numbers indicate the number of nested calls to set_waiting_cursor() */
+	guint cursor_ref;
+
 	/* This is a scratch list for when we're removing files */
 	GList *list;
 	guint current_to_be_removed : 1;
@@ -216,21 +223,17 @@ totem_playlist_get_toplevel (TotemPlaylist *playlist)
 }
 
 static void
-totem_playlist_set_waiting_cursor (TotemPlaylist *playlist)
+set_waiting_cursor (TotemPlaylist *playlist)
 {
-	GtkWidget *parent;
-
-	parent = GTK_WIDGET (totem_playlist_get_toplevel (playlist));
-	totem_gdk_window_set_waiting_cursor (gtk_widget_get_window (parent));
+	totem_gdk_window_set_waiting_cursor (gtk_widget_get_window (GTK_WIDGET (totem_playlist_get_toplevel (playlist))));
+	playlist->priv->cursor_ref++;
 }
 
 static void
-totem_playlist_unset_waiting_cursor (TotemPlaylist *playlist)
+unset_waiting_cursor (TotemPlaylist *playlist)
 {
-	GtkWidget *parent;
-
-	parent = GTK_WIDGET (totem_playlist_get_toplevel (playlist));
-	gdk_window_set_cursor (gtk_widget_get_window (parent), NULL);
+	if (--playlist->priv->cursor_ref < 1)
+		gdk_window_set_cursor (gtk_widget_get_window (GTK_WIDGET (totem_playlist_get_toplevel (playlist))), NULL);
 }
 
 static void
@@ -389,12 +392,10 @@ totem_playlist_mrl_to_title (const gchar *mrl)
 static void
 totem_playlist_update_save_button (TotemPlaylist *playlist)
 {
-	GtkWidget *button;
 	gboolean state;
 
-	button = GTK_WIDGET (gtk_builder_get_object (playlist->priv->xml, "save_button"));
 	state = (!playlist->priv->disable_save_to_disk) && (PL_LEN != 0);
-	gtk_widget_set_sensitive (button, state);
+	gtk_widget_set_sensitive (playlist->priv->save_button, state);
 }
 
 static gboolean
@@ -485,6 +486,15 @@ gtk_tree_selection_has_selected (GtkTreeSelection *selection)
 	return retval;
 }
 
+static void
+drop_finished_cb (TotemPlaylist *playlist, GAsyncResult *result, gpointer user_data)
+{
+	/* Emit the "changed" signal once the last dropped MRL has been added to the playlist */
+	g_signal_emit (G_OBJECT (playlist),
+	               totem_playlist_table_signals[CHANGED], 0,
+	               NULL);
+}
+
 static void
 drop_cb (GtkWidget        *widget,
          GdkDragContext   *context, 
@@ -498,16 +508,19 @@ drop_cb (GtkWidget        *widget,
 	char **list;
 	GList *p, *file_list;
 	guint i;
+	GdkDragAction action;
 
-	if (context->suggested_action == GDK_ACTION_ASK) {
-		context->action = totem_drag_ask (PL_LEN != 0);
-		if (context->action == GDK_ACTION_DEFAULT) {
+	if (gdk_drag_context_get_suggested_action (context) == GDK_ACTION_ASK) {
+		action = totem_drag_ask (PL_LEN != 0);
+		gdk_drag_status (context, action, GDK_CURRENT_TIME);
+		if (action == GDK_ACTION_DEFAULT) {
 			gtk_drag_finish (context, FALSE, FALSE, _time);
 			return;
 		}
 	}
 
-	if (context->action == GDK_ACTION_MOVE)
+	action = gdk_drag_context_get_selected_action (context);
+	if (action == GDK_ACTION_MOVE)
 		totem_playlist_clear (playlist);
 
 	list = g_uri_list_extract_uris ((char *) gtk_selection_data_get_data (data));
@@ -525,8 +538,6 @@ drop_cb (GtkWidget        *widget,
 		return;
 	}
 
-	totem_playlist_set_waiting_cursor (playlist);
-
 	playlist->priv->tree_path = gtk_tree_path_new ();
 	gtk_tree_view_get_dest_row_at_pos (GTK_TREE_VIEW (playlist->priv->treeview),
 					   x, y,
@@ -559,7 +570,12 @@ drop_cb (GtkWidget        *widget,
 			}
 		}
 
-		totem_playlist_add_mrl (playlist, filename, title);
+		/* Add the MRL to the playlist asynchronously. If it's the last MRL, emit the "changed"
+		 * signal once we're done adding it */
+		if (p->next == NULL)
+			totem_playlist_add_mrl (playlist, filename, title, TRUE, NULL, (GAsyncReadyCallback) drop_finished_cb, NULL);
+		else
+			totem_playlist_add_mrl (playlist, filename, title, TRUE, NULL, NULL, NULL);
 
 		g_free (filename);
 	}
@@ -569,12 +585,6 @@ drop_cb (GtkWidget        *widget,
 	gtk_drag_finish (context, TRUE, FALSE, _time);
 	gtk_tree_path_free (playlist->priv->tree_path);
 	playlist->priv->tree_path = NULL;
-
-	totem_playlist_unset_waiting_cursor (playlist);
-
-	g_signal_emit (G_OBJECT (playlist),
-			totem_playlist_table_signals[CHANGED], 0,
-			NULL);
 }
 
 void
@@ -797,23 +807,16 @@ drag_end_cb (GtkWidget *treeview, GdkDragContext *context, gpointer data)
 static void
 selection_changed (GtkTreeSelection *treeselection, TotemPlaylist *playlist)
 {
-	GtkWidget *remove_button, *up_button, *down_button;
 	gboolean sensitivity;
 
-	remove_button = GTK_WIDGET (gtk_builder_get_object (playlist->priv->xml,
-			"remove_button"));
-	up_button = GTK_WIDGET (gtk_builder_get_object (playlist->priv->xml, "up_button"));
-	down_button = GTK_WIDGET (gtk_builder_get_object (playlist->priv->xml,
-			"down_button"));
-
 	if (gtk_tree_selection_has_selected (treeselection))
 		sensitivity = TRUE;
 	else
 		sensitivity = FALSE;
 
-	gtk_widget_set_sensitive (remove_button, sensitivity);
-	gtk_widget_set_sensitive (up_button, sensitivity);
-	gtk_widget_set_sensitive (down_button, sensitivity);
+	gtk_widget_set_sensitive (playlist->priv->remove_button, sensitivity);
+	gtk_widget_set_sensitive (playlist->priv->up_button, sensitivity);
+	gtk_widget_set_sensitive (playlist->priv->down_button, sensitivity);
 }
 
 /* This function checks if the current item is NULL, and try to update it
@@ -851,22 +854,18 @@ totem_playlist_add_files (GtkWidget *widget, TotemPlaylist *playlist)
 {
 	GSList *filenames, *l;
 
-	filenames = totem_add_files (totem_playlist_get_toplevel (playlist),
-			NULL);
+	filenames = totem_add_files (totem_playlist_get_toplevel (playlist), NULL);
 	if (filenames == NULL)
 		return;
 
-	totem_playlist_set_waiting_cursor (playlist);
-
 	for (l = filenames; l != NULL; l = l->next) {
 		char *mrl;
 
 		mrl = l->data;
-		totem_playlist_add_mrl (playlist, mrl, NULL);
+		totem_playlist_add_mrl (playlist, mrl, NULL, TRUE, NULL, NULL, NULL);
 		g_free (mrl);
 	}
 
-	totem_playlist_unset_waiting_cursor (playlist);
 	g_slist_free (filenames);
 }
 
@@ -1633,17 +1632,20 @@ totem_playlist_dispose (GObject *object)
 {
 	TotemPlaylist *playlist = TOTEM_PLAYLIST (object);
 
-	if (playlist->priv->parser == NULL)
-		return;
+	if (playlist->priv->parser != NULL) {
+		g_object_unref (playlist->priv->parser);
+		playlist->priv->parser = NULL;
+	}
 
-	g_object_unref (playlist->priv->parser);
-	playlist->priv->parser = NULL;
-
-	if (playlist->priv->ui_manager != NULL)
+	if (playlist->priv->ui_manager != NULL) {
 		g_object_unref (G_OBJECT (playlist->priv->ui_manager));
+		playlist->priv->ui_manager = NULL;
+	}
 
-	if (playlist->priv->action_group != NULL)
+	if (playlist->priv->action_group != NULL) {
 		g_object_unref (G_OBJECT (playlist->priv->action_group));
+		playlist->priv->action_group = NULL;
+	}
 
 	G_OBJECT_CLASS (totem_playlist_parent_class)->dispose (object);
 }
@@ -1663,6 +1665,7 @@ static void
 totem_playlist_init (TotemPlaylist *playlist)
 {
 	GtkWidget *container;
+	GtkBuilder *xml;
 
 	playlist->priv = G_TYPE_INSTANCE_GET_PRIVATE (playlist, TOTEM_TYPE_PLAYLIST, TotemPlaylistPrivate);
 	playlist->priv->parser = totem_pl_parser_new ();
@@ -1679,21 +1682,29 @@ totem_playlist_init (TotemPlaylist *playlist)
 			G_CALLBACK (totem_playlist_entry_parsed),
 			playlist);
 
-	playlist->priv->xml = totem_interface_load ("playlist.ui", TRUE, NULL, playlist);
+	xml = totem_interface_load ("playlist.ui", TRUE, NULL, playlist);
 
-	if (playlist->priv->xml == NULL)
+	if (xml == NULL)
 		return;
 
 	/* popup menu */
-	playlist->priv->action_group = GTK_ACTION_GROUP (gtk_builder_get_object (playlist->priv->xml, "playlist-action-group"));
-	playlist->priv->ui_manager = GTK_UI_MANAGER (gtk_builder_get_object (playlist->priv->xml, "totem-playlist-ui-manager"));
-	
+	playlist->priv->action_group = GTK_ACTION_GROUP (gtk_builder_get_object (xml, "playlist-action-group"));
+	g_object_ref (playlist->priv->action_group);
+	playlist->priv->ui_manager = GTK_UI_MANAGER (gtk_builder_get_object (xml, "totem-playlist-ui-manager"));
+	g_object_ref (playlist->priv->ui_manager);
+
 	gtk_widget_add_events (GTK_WIDGET (playlist), GDK_KEY_PRESS_MASK);
 	g_signal_connect (G_OBJECT (playlist), "key_press_event",
 			  G_CALLBACK (totem_playlist_key_press), playlist);
 
+	/* Buttons */
+	playlist->priv->save_button = GTK_WIDGET (gtk_builder_get_object (xml, "save_button"));;
+	playlist->priv->remove_button = GTK_WIDGET (gtk_builder_get_object (xml, "remove_button"));
+	playlist->priv->up_button = GTK_WIDGET (gtk_builder_get_object (xml, "up_button"));
+	playlist->priv->down_button = GTK_WIDGET (gtk_builder_get_object (xml, "down_button"));
+
 	/* Reparent the vbox */
-	container = GTK_WIDGET (gtk_builder_get_object (playlist->priv->xml, "vbox4"));
+	container = GTK_WIDGET (gtk_builder_get_object (xml, "vbox4"));
 	g_object_ref (container);
 	gtk_box_pack_start (GTK_BOX (playlist),
 			container,
@@ -1702,8 +1713,7 @@ totem_playlist_init (TotemPlaylist *playlist)
 			0);         /* padding */
 	g_object_unref (container);
 
-	playlist->priv->treeview = GTK_WIDGET (gtk_builder_get_object
-		(playlist->priv->xml, "treeview1"));
+	playlist->priv->treeview = GTK_WIDGET (gtk_builder_get_object (xml, "treeview1"));
 	init_treeview (playlist->priv->treeview, playlist);
 	playlist->priv->model = gtk_tree_view_get_model
 		(GTK_TREE_VIEW (playlist->priv->treeview));
@@ -1716,6 +1726,8 @@ totem_playlist_init (TotemPlaylist *playlist)
 	init_config (playlist);
 
 	gtk_widget_show_all (GTK_WIDGET (playlist));
+
+	g_object_unref (xml);
 }
 
 GtkWidget*
@@ -1724,7 +1736,7 @@ totem_playlist_new (void)
 	TotemPlaylist *playlist;
 
 	playlist = TOTEM_PLAYLIST (g_object_new (TOTEM_TYPE_PLAYLIST, NULL));
-	if (playlist->priv->xml == NULL || playlist->priv->ui_manager == NULL) {
+	if (playlist->priv->ui_manager == NULL) {
 		g_object_unref (playlist);
 		return NULL;
 	}
@@ -1822,38 +1834,36 @@ totem_playlist_add_one_mrl (TotemPlaylist *playlist,
 	return TRUE;
 }
 
-gboolean
-totem_playlist_add_mrl_with_cursor (TotemPlaylist *playlist, const char *mrl,
-				    const char *display_name)
+typedef struct {
+	GAsyncReadyCallback callback;
+	gpointer user_data;
+	gboolean cursor;
+	TotemPlaylist *playlist;
+	gchar *mrl;
+	gchar *display_name;
+} AddMrlData;
+
+static void
+add_mrl_data_free (AddMrlData *data)
 {
-	gboolean retval;
-
-	totem_playlist_set_waiting_cursor (playlist);
-	retval = totem_playlist_add_mrl (playlist, mrl, display_name);
-	totem_playlist_unset_waiting_cursor (playlist);
-
-	return retval;
+	g_object_unref (data->playlist);
+	g_free (data->mrl);
+	g_free (data->display_name);
+	g_slice_free (AddMrlData, data);
 }
 
-gboolean
-totem_playlist_add_mrl (TotemPlaylist *playlist, const char *mrl,
-			const char *display_name)
+static gboolean
+handle_parse_result (TotemPlParserResult res, TotemPlaylist *playlist, const gchar *mrl, const gchar *display_name)
 {
-	TotemPlParserResult res;
-
-	g_return_val_if_fail (mrl != NULL, FALSE);
-
-	res = totem_pl_parser_parse (playlist->priv->parser, mrl, FALSE);
-
 	if (res == TOTEM_PL_PARSER_RESULT_UNHANDLED)
 		return totem_playlist_add_one_mrl (playlist, mrl, display_name);
-	if (res == TOTEM_PL_PARSER_RESULT_ERROR)
-	{
+	if (res == TOTEM_PL_PARSER_RESULT_ERROR) {
 		char *msg;
 
-		msg = g_strdup_printf (_("The playlist '%s' could not be parsed, it might be damaged."), display_name ? display_name : mrl);
+		msg = g_strdup_printf (_("The playlist '%s' could not be parsed. It might be damaged."), display_name ? display_name : mrl);
 		totem_playlist_error (_("Playlist error"), msg, playlist);
 		g_free (msg);
+
 		return FALSE;
 	}
 	if (res == TOTEM_PL_PARSER_RESULT_IGNORED)
@@ -1862,6 +1872,79 @@ totem_playlist_add_mrl (TotemPlaylist *playlist, const char *mrl,
 	return TRUE;
 }
 
+static void
+add_mrl_cb (TotemPlParser *parser, GAsyncResult *result, AddMrlData *data)
+{
+	TotemPlParserResult res;
+	GSimpleAsyncResult *async_result;
+	GError *error = NULL;
+
+	/* Finish parsing the playlist */
+	res = totem_pl_parser_parse_finish (parser, result, &error);
+
+	/* Remove the cursor, if one was set */
+	if (data->cursor)
+		unset_waiting_cursor (data->playlist);
+
+	/* Create an async result which will return the result to the code which called totem_playlist_add_mrl() */
+	if (error != NULL)
+		async_result = g_simple_async_result_new_from_error (G_OBJECT (data->playlist), data->callback, data->user_data, error);
+	else
+		async_result = g_simple_async_result_new (G_OBJECT (data->playlist), data->callback, data->user_data, totem_playlist_add_mrl);
+
+	/* Handle the various return cases from the playlist parser */
+	g_simple_async_result_set_op_res_gboolean (async_result, handle_parse_result (res, data->playlist, data->mrl, data->display_name));
+
+	/* Free the closure's data, now that we're finished with it */
+	add_mrl_data_free (data);
+
+	/* Synchronously call the calling code's callback function (i.e. what was passed to totem_playlist_add_mrl()'s @callback parameter)
+	 * in the main thread to return the result */
+	g_simple_async_result_complete (async_result);
+}
+
+void
+totem_playlist_add_mrl (TotemPlaylist *playlist, const char *mrl, const char *display_name, gboolean cursor,
+                        GCancellable *cancellable, GAsyncReadyCallback callback, gpointer user_data)
+{
+	AddMrlData *data;
+
+	g_return_if_fail (mrl != NULL);
+
+	/* Display a waiting cursor if required */
+	if (cursor)
+		set_waiting_cursor (playlist);
+
+	/* Build the data struct to pass to the callback function */
+	data = g_slice_new (AddMrlData);
+	data->callback = callback;
+	data->user_data = user_data;
+	data->cursor = cursor;
+	data->playlist = g_object_ref (playlist);
+	data->mrl = g_strdup (mrl);
+	data->display_name = g_strdup (display_name);
+
+	/* Start parsing the playlist. Once this is complete, add_mrl_cb() is called, which will interpret the results and call @callback to
+	 * finish the process. */
+	totem_pl_parser_parse_async (playlist->priv->parser, mrl, FALSE, cancellable, (GAsyncReadyCallback) add_mrl_cb, data);
+}
+
+gboolean
+totem_playlist_add_mrl_finish (TotemPlaylist *playlist, GAsyncResult *result)
+{
+	g_assert (g_simple_async_result_get_source_tag (G_SIMPLE_ASYNC_RESULT (result)) == totem_playlist_add_mrl);
+
+	return g_simple_async_result_get_op_res_gboolean (G_SIMPLE_ASYNC_RESULT (result));
+}
+
+gboolean
+totem_playlist_add_mrl_sync (TotemPlaylist *playlist, const char *mrl, const char *display_name)
+{
+	g_return_val_if_fail (mrl != NULL, FALSE);
+
+	return handle_parse_result (totem_pl_parser_parse (playlist->priv->parser, mrl, FALSE), playlist, mrl, display_name);
+}
+
 static gboolean
 totem_playlist_clear_cb (GtkTreeModel *model,
 			 GtkTreePath  *path,
